(***********************************************************************)
(*                                                                     *)
(*                       Ambients in JoCaml                            *)
(*                                                                     *)
(*               Alan Schmitt, projet Para, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1999 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Syntax
open Hashtbl
open Print
  
type id_t = (int * (string * int))
      
and sub_in_t = << id_t * string * <<unit>> * int >>
and sub_out_t = << id_t * string * <<unit>> * int >>
    
and a_in_t = << string * << unit >> >>
and a_out_t = << string * << unit >> >>
and a_open_t = << string * << unit >> >>
and comm_in_t = << << capacity >> >>
and comm_out_t = << capacity * << unit >> >>
    
and commit_t = << subambient_interface * << unit >> * int >>
and opening_t = << subambient_interface * a_in_t * a_out_t * a_open_t * comm_in_t * comm_out_t >>
and sub_t = << id_t * subambient_interface * string * commit_t * opening_t * << << subambient_interface * id_t  >> * subambient_interface * id_t >> * <<unit>> >>
    
and subambient_interface = {there: Join.location; sub: sub_t; sub_in: sub_in_t; sub_out: sub_out_t}
and ambient_interface = {e: subambient_interface; a_in: a_in_t; a_out: a_out_t; a_open: a_open_t; comm_in: comm_in_t; comm_out: comm_out_t}
      
and sub_by_id = {ee: subambient_interface; name: string; commit: commit_t; opening: opening_t; bang: <<unit>>}
      
exception Amb_has_moved
exception Action_impossible
exception Amb_syntax_error
    
(* the following type is used for storing messages sent
   (as in the log) and for storing active action messages
*)
type log_type = IN of (id_t * string * <<unit>> * int)
  | OUT of (id_t * string * <<unit>> * int)
  | OPEN of (string * <<unit>>)
	
let last_id = ref 0
let last_new_id = ref 0
    
let rec trad f env bang n = function
    Par (p, q) -> spawn { {trad f env bang n p ;} | {trad f env bang n q ;} }
  | Seq (capa, p) ->
      begin
	match capa with
	  Seq_capa (c1, c2) -> trad f env bang n (Seq (c1, (Seq (c2, p))))
	| In name ->
	    begin
	      let after_replace = replace_from_env name env in
	      let name' = match after_replace with
		Ident ident -> ident
	      |	_ -> raise Amb_syntax_error
	      in  
	      match bang with
		None ->  
		  let def k! () = {trad f env None n p ;} in
		  print_string ("in "^name'^"."); print_process p; print_string " was translated\n"; flush stdout;
		  spawn {f.a_in(name', k)}
	      | Some bp ->
		  let def k! () = {trad f env None n p ;} | bp () in
		  print_string ("in "^name'^"."); print_process p; print_string " was translated under bang\n"; flush stdout;
		  spawn {f.a_in(name', k)}
	    end
	| Out name ->
	    begin
	      let after_replace = replace_from_env name env in
	      let name' = match after_replace with
		Ident ident -> ident
	      |	_ -> raise Amb_syntax_error
	      in
	      match bang with
		None ->  
		  let def k! () = {trad f env None n p ;} in
		  print_string ("out "^name'^"."); print_process p; print_string " was translated\n"; flush stdout;
		  spawn {f.a_out(name', k)}
	      | Some bp ->
		  let def k! () = {trad f env None n p ;} | bp () in
		  print_string ("out "^name'^"."); print_process p; print_string " was translated under bang\n"; flush stdout;
		  spawn {f.a_out(name', k)}
	    end
	| Open name ->
	    begin
	      let after_replace = replace_from_env name env in
	      let name' = match after_replace with
		Ident ident -> ident
	      |	_ -> raise Amb_syntax_error
	      in
	      match bang with
		None ->  
		  let def k! () = {trad f env None n p ;} in
		  print_string ("open "^name'^"."); print_process p; print_string " was translated\n"; flush stdout;
		  spawn {f.a_open(name', k)}
	      | Some bp ->
		  let def k! () = {trad f env None n p ;} | bp () in
		  print_string ("open "^name'^"."); print_process p; print_string " was translated under bang\n"; flush stdout;
		  spawn {f.a_open(name', k)}
	    end
	| Ident name ->
	    begin
	      let after_replace = replace_from_env name env in
	      match after_replace with
		Ident ident -> raise Amb_syntax_error
	      |	Seq_capa (c1, c2) -> if (is_first_capa_ident c1) then (raise Amb_syntax_error) else (trad f env bang n (Seq (c1, Seq (c2, p))))
	      |	_ -> trad f env bang n (Seq (after_replace, p))
	    end
      end
  | Amb (name, p, b) ->
      let after_replace = replace_from_env name env in
      let name' = match after_replace with
	Ident ident -> ident
      |	_ -> raise Amb_syntax_error
      in
      print_string (name'^"["); print_process p; print_string "] is being translated\n"; flush stdout;
      ambient name' f.e env p b bang n
  | New (name, p) ->
      trad f ( (name, (make_new_id n)) :: env ) bang n p
  | Comm_in (avar, p) ->
      begin
	match bang with
	  None ->  
	    let def k! value = {trad f ((avar, value) :: env) None n p ;} in
	    print_string ("("^avar^")."); print_process p; print_string " was translated\n"; flush stdout;
	    spawn {f.comm_in (k)}
	| Some bp ->  
	    let def k! value = {trad f ((avar, value) :: env) None n p ;} | bp () in
	    print_string ("("^avar^")."); print_process p; print_string " was translated under bang\n"; flush stdout;
	    spawn {f.comm_in (k)}
      end
  | Comm_out c ->
      begin
	let val_to_send =
	  begin
	    match c with
	      Ident c' -> replace_from_env c' env
	    | _ as c' -> c'
	  end in
	match bang with
	  None ->
	    let def k! () = {} in
	    print_string "<"; print_capa val_to_send; print_string "> was translated\n"; flush stdout;
	    spawn {f.comm_out (val_to_send, k)}
	| Some bp ->
	    let def k! () = bp () in
	    print_string "<"; print_capa val_to_send; print_string "> was translated under bang\n"; flush stdout;
	    spawn {f.comm_out (val_to_send, k)}
      end
  | Bang p ->
      let def bp! () = {trad f env (Some bp) n p;} in
      spawn { bp () | bp () }
  | Zero -> print_string "end of process\n"; flush stdout
	
and ambient name initial_e env code_to_run reg bang n =
  let loc here
  do {
    let h_sub_by_id = create 100 in
    let h_sub_by_name = create 100 in
    let local_log = ref [] in
    let active_messages = ref [] in
    let last_index_for_message = ref 0 in
    let isbang = ref bang in
    
    (* un message sub est detecte et range dans les tables de hashage.
       On en profite pour verifier si on ne peut pas consommer une
       action en attente et on active la continuation *)
    let def live! (e', id') | sub! (id, e, sname, scommit, sopening, cont, tbang) =
      add h_sub_by_id id {ee = e; name = sname; commit = scommit; opening = sopening; bang = tbang};
      add_elt h_sub_by_name sname id;
      print_string ("amb "^name^" "); print_id id'; print_string (" has new subambient "^sname^" "); print_id id; print_newline ();
      print_string ("printing active messages for ambient "^name^" "); print_id id'; print_newline ();
      print_active_messages !active_messages;
      flush stdout;
      let this_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
      active_messages := (consume_actions h_sub_by_id h_sub_by_name name this_e e' a_in a_out a_open comm_in comm_out test_bang !active_messages);
      { cont (live, e', id') }
	
    (* un message in ou out est emis dans cet ambient. On le met dans
       le log et on emet un sub_in ou un sub_out dans l'amient pere *)
    or live! (e, id) | a_in! (a, k) =
      let v = !last_index_for_message in
      print_string ("in "^a^" message by "^name^", "); print_id id; print_string (" is being added to log with index "^(string_of_int v)^"\n"); flush stdout;
      (local_log := (IN((0,("",0)), a, k, v) :: !local_log));
      incr last_index_for_message;
      live (e, id) | e.sub_in(id, a, k, v) 
	
    or live! (e, id) | a_out! (a, k) =
      let v = !last_index_for_message in
      (local_log := (OUT((0,("",0)), a, k, v) :: !local_log));
      incr last_index_for_message;
      live (e, id) | e.sub_out(id, a, k, v)
	
    (* un message sub_in, sub_out ou open est vivant dans cet ambient.
       On essaie de le consommer. Si on ne peut pas, on le garde pour plus
       tard (ie arrivee d'un nouveau subambient ou ouverture de celui-ci *)
    
    or live! (e, id') | sub_in! (id, a, k, v) = print_string ("in "^a^" message by "); print_id id; print_string (" is considerered for being added to active message list while being in amb "^name^" "); print_id id'; print_newline (); flush stdout;
      let this_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
      if (not (use_message h_sub_by_id h_sub_by_name name this_e e a_in a_out a_open comm_in comm_out test_bang (IN (id, a, k, v)))) then
	{{ (active_messages := (IN(id, a, k, v) :: !active_messages));
	   print_string ("in "^a^" by "); print_id id; print_string " did not work\n"; flush stdout; } | live(e, id')}
      else live(e, id')

    or live! (e, id') | sub_out! (id, a, k, v) = print_string ("out "^a^" message by "); print_id id; print_string (" is considered for being added to active message list while being in amb "^name^" "); print_id id'; print_newline (); flush stdout;
      let this_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
      if (not (use_message h_sub_by_id h_sub_by_name name this_e e a_in a_out a_open comm_in comm_out test_bang (OUT (id, a, k, v))))
      then
	{{ (active_messages := (OUT(id, a, k, v) :: !active_messages));
	   print_string ("out "^a^" by "); print_id id; print_string " did not work\n"; flush stdout; } | live(e, id')}
      else live(e, id')

    or live! (e, id') | a_open! (a, k) = print_string ("open "^a^" message is considered for being added to active message list while being in amb "^name^" "); print_id id'; print_newline (); flush stdout;
       let this_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
       if (not (use_message h_sub_by_id h_sub_by_name name this_e e a_in a_out a_open comm_in comm_out test_bang (OPEN (a, k)))) then
	 {{ (active_messages := (OPEN(a, k) :: !active_messages));
	    print_string ("open "^a^" did not work\n"); flush stdout; } | live (e, id')}
       else live(e, id')
	   
    (* la communication. la continuation pour le comm_out ne sert que si il
       y a un bang ... *)
    or live!(e, id) | comm_in! k | comm_out! (v, k') = live(e, id) | k(v) | k' ()
       
    (* le commit, en flushant le log pour que le nouveau parent connaisse
       les messages actifs *)
    or live! (e, id) | commit! (e', k, v) =
      go(e'.there);
      print_string ("before removing "^(string_of_int v)^": itering local log for "^name^" "); print_id id; print_string ":\n";
      print_active_messages !local_log; flush stdout;
      local_log := remove_from_log v !local_log;
       print_string ("after removing "^(string_of_int v)^": itering local log for "^name^" "); print_id id; print_string ":\n"; 
       print_active_messages !local_log; flush stdout;
      let id' = next_id id in
       let new_e = {there=here; sub=sub; sub_in=sub_in; sub_out=sub_out} in
       (* Cette continuation permet de s'assurer que le message sub
          a bien t enregistr dans la table de hashage avant de
          rmettre les messages en attente. On reoit le message live
          pour s'assurer que aucun message ne sera ajout dans
          le log pendant l'numration *)
       let def cont!(l, e, i) = {
         let log_to_iter = !local_log in
         local_log:= [];
         iter_log a_in a_out log_to_iter;
         spawn {l(e, i) | k ()};
       } in
       spawn {live (e', id') | e'.sub(id', new_e, name, commit, opening, cont, test_bang)};
       

      or live! (e, id) | opening! (e', par_in, par_out, par_open, par_comm_in, par_comm_out) =
       (* go (e'.there); *)
       {opened(e', par_in, par_out, par_open, par_comm_in, par_comm_out)
       |
       (* on met le log local (ie, les in et out qui ne seront plus valables
          par ce que l'on n'existera plus *)
       {let log_to_iter = !local_log in
       local_log:= [];
       iter_log par_in par_out log_to_iter;}
       | {
       (* on met les actions locales en attente *)
       let l = !active_messages in
       active_messages := [];
       flush_actions sub_in sub_out a_open l;
       }
       | {
       (* on met les sun locaux et on vide les tables *)
       flush_sub_table h_sub_by_id sub;
       clear h_sub_by_name;
       }
       }

     (* on est opened, donc on forward les actions *)
       
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | a_in!(a, k) = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | par_in(a, k)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | a_out!(a, k) = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | par_out(a, k)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | a_open!(a, k) = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | par_open(a, k)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | comm_in! k = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | par_comm_in(k)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | comm_out! (c, k) = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | par_comm_out (c, k)  
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | sub! (id, e', sname, scommit, sopening, c, b) = {print_string ("sub message with name "^sname^" and id "); print_id id; print_string (" is forwarded by opened ambient "^name^"\n"); flush stdout; } | opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | e.sub (id, e', sname, scommit, sopening, c, b)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | sub_in!(id, a, k, v) = {print_string ("in "^a^" message by amb "); print_id id; print_string (" forwarded from opened ambient "^name^"\n"); flush stdout; } | opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | e.sub_in(id, a, k, v)
      or opened!(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | sub_out!(id, a, k, v) = opened(e, par_in, par_out, par_open, par_comm_in, par_comm_out) | e.sub_out(id, a, k, v)
      
     (* ce canal sert  grer les bang ... *)
     or test_bang! () =
       match (!isbang) with
         None -> {}
       | Some bp -> { isbang := None; spawn { bp () };}
     
     in {
       let id=make_id n in
       let e = {there = here; sub = sub; sub_in = sub_in; sub_out = sub_out} in
       let f = {e = e; a_in = a_in; a_out = a_out; a_open = a_open; comm_in = comm_in; comm_out = comm_out } in
       let def foo! (l, e, i) = l(e, i) | {trad f env None n code_to_run ;} in
       spawn {initial_e.sub(id, e, name, commit, opening, foo, test_bang)};
       print_string ("amb initialised: name: "^name^", id: "); print_id id; print_string " with code: "; print_process code_to_run; print_newline (); flush stdout;
       (if reg then (
       Ns.register name f vartype;
       print_string (name^" registered\n"); flush stdout));
       {live (initial_e, id)}
       }
  }
  in ()

(* une fonction qui reemet tous les messages sub d'une table en les
   supprimant *)
(* on a une continuation qui est pipo car cette fonction n'est
   appele que quand on fait un open *)
and flush_sub_table h sub =
  let def foo! (l, e, i) = {} in
  iter (fun id v -> spawn {sub(id, v.ee, v.name, v.commit, v.opening, foo, v.bang)}) h;
  clear h
      
(* les fonctions suivantes reemettent les messages actifs contenus
   dans la liste des messages actifs *)
and flush_actions sub_in sub_out a_open = function
    [] -> ()
  | msg::l -> match msg with
      IN (id, a, k, v) -> spawn {sub_in (id, a, k, v)}; flush_actions sub_in sub_out a_open l
    | OUT (id, a, k, v) -> spawn {sub_out (id, a, k, v)}; flush_actions sub_in sub_out a_open l
    | OPEN (a, k) -> spawn {a_open (a, k)}; flush_actions sub_in sub_out a_open l

(* la fonction suivante cherche les actions que l'on peut satisfaire
   en supprimant celles qui ne sont pas satisfiables ou qui n'ont plus
   lieu d'etre *)
and consume_actions hid hname name this_e e a_in a_out a_open comm_in comm_out test_bang = function
    [] -> []
  | msg::l -> if (use_message hid hname name this_e e a_in a_out a_open comm_in comm_out test_bang msg) then
      (consume_actions hid hname name this_e e a_in a_out a_open comm_in comm_out test_bang l) else
      msg :: (consume_actions hid hname name this_e e a_in a_out a_open comm_in comm_out test_bang l)
      
(* semantique: si le message est utilise (soit parce que
   l'action a eu lieu, soit par ce que le sous-ambient
   n'existe plus), retourne true. Retourne false sinon *)
    
and use_message hid hname name this_e e a_in a_out a_open comm_in comm_out test_bang = function
    IN (id, a, k, v) -> use_in_message hid hname id a k v
  | OUT (id, a, k, v) -> use_out_message hid hname name e id a k v test_bang
  | OPEN (a, k) -> use_open_message hid hname this_e a_in a_out a_open comm_in comm_out a k
	
and use_in_message hid hname id a k v =
  try
    let amb1 = (
      try
	(* on commence par verifier que le sous ambient qui a lance
	   le message existe encore *)
	find hid id
      with Not_found -> raise Amb_has_moved ) in
    (* on enleve amb1 pour etre sur de ne pas le rechoisir
       si il est du genre: n[ in n ] *)
    if (amb1.name = a) then remove_elt hname a id;
    let id2 = (
      try
	(* on cherche l'ambient dans lequel on veut aller *)
	find_elt hname a
      with Not_found -> raise Action_impossible ) in
    (* les deux ambient sont present, on agit *)
    (* on commence par aller chercher l'ambient correspondant *)
    let amb2 = find hid id2 in
    print_string ("amb "^amb1.name^" with id "); print_id id; print_string (" is moving in amb "^amb2.name^" with id "); print_id id2; print_newline(); flush stdout;
    (* on supprime le "message" pour l'ambient 1 *)
    remove hid id;
    if (amb1.name != a) then remove_elt hname amb1.name id;
    spawn {amb1.commit(amb2.ee, k, v) | amb1.bang () | amb2.bang ()};
    true
    
  with
    Amb_has_moved -> print_string ("*** message in "^a^" by amb "); print_id id; print_string " aborted: the amb has moved ***\n"; flush stdout; true
  | Action_impossible -> false
      
and use_out_message hid hname name e id a k v test_bang =
  try
    let amb1 = (
      try
	(* on commence par verifier que le sous ambient qui a lance
	   le message existe encore *)
	find hid id
      with Not_found -> raise Amb_has_moved ) in
    if (not (name = a)) then (raise Action_impossible);
    (* tout est ok pour sortir de l'ambient englobant *)
    print_string ("amb "^amb1.name^" with id "); print_id id; print_string (" is moving out of amb "^name^"\n"); flush stdout;
    remove hid id;
    remove_elt hname amb1.name id;
    spawn {amb1.commit(e, k, v) | amb1.bang () | test_bang ()};
    true
      
  with
    Amb_has_moved -> print_string ("*** message out "^a^" by amb "); print_id id; print_string " aborted: the amb has moved ***\n"; flush stdout; true
  | Action_impossible -> false
    
and use_open_message hid hname e a_in a_out a_open comm_in comm_out a k =
  try
    let id_a = (
      try
	find_elt hname a
      with Not_found -> raise Action_impossible ) in
    let amb2 = find hid id_a in
    print_string ("amb "^amb2.name^" with id "); print_id id_a; print_string " is being opened\n";
    flush stdout;
    (* le open peut avoir lieu on commence par enlever l'ambient *)
    remove_elt hname a id_a;
    remove hid id_a;
    (* on lance un message opening et la continuation *)
    spawn {amb2.opening (e, a_in, a_out, a_open, comm_in, comm_out ) | k () | amb2.bang ()};
    true
      
  with
    Action_impossible -> false
	
(* les fonctions suivantes servent a gerer la table
   ou est associe un nom d'ambient a la liste des
   identificateurs qui portent ce nom *)

and find_elt hname a =
  match (find hname a) with
    [] -> raise Not_found
  | id :: l -> id

and remove_elt_from_list id = function
    [] -> raise Not_found
  | id' :: l -> if (id = id') then l
	else id' :: (remove_elt_from_list id l)

and remove_elt hname a id =
  let l = find hname a in
  remove hname a;
  add hname a (remove_elt_from_list id l)

and add_elt hname a id =
  let l =
    try
      let temp = find hname a in
      remove hname a;
      temp
    with Not_found -> []
  in
  add hname a (id :: l)

(* on ne gre pas le cas open car il n'y en a pas dans le log local *)
    
and remove_from_log cont = function
    [] -> raise Not_found
  | a :: l -> match a with
      IN (_, _, _, v) -> if (v=cont) then l else (a :: (remove_from_log cont l))
    | OUT (_, _, _, v) -> if (v=cont) then l else (a :: (remove_from_log cont l))
    
(* Cette fonction cree de nouveau identificateurs *)						    
and make_id n =
    incr last_id;
    (n, ("id"^(string_of_int !last_id), 0))

and next_id id =
  ((fst id), (fst(snd id), (snd (snd id))+1))

and print_id id = print_string ((fst (snd id))^"_"^(string_of_int (fst id))^", "^(string_of_int (snd (snd id))))
       
(* Cette fonction cre de nouveaux identificateurs pour les new *)

and make_new_id n =
    incr last_new_id;
    Ident ("*new_"^(string_of_int n)^","^(string_of_int !last_new_id))
      
(* Cette fonction reemet tous les messages stockes dans le log
   Rq: il ne peut pas y avoir de message "open" dans le log, ils n'y
   sont jamais mis*)    
and iter_log a_in a_out = function
    [] -> ()
  | a::l -> match a with
      IN (_, a, k, _) -> spawn {a_in (a, k)}; print_string "in spawned\n"; flush stdout; iter_log a_in a_out l
    | OUT (_, a, k, _) -> spawn {a_out (a, k)}; iter_log a_in a_out l
    
(* cette fonction imprime les messages actifs en attente *)
and print_active_messages = function
    [] -> print_string "-end of list-\n"
  | a::l -> match a with
      IN (id, a, k, v) -> print_string ((string_of_int v)^" in "^a^" by amb "); print_id id; print_newline (); print_active_messages l
    | OUT (id, a, k, v) -> print_string ((string_of_int v)^" out "^a^" by amb "); print_id id; print_newline (); print_active_messages l
    | OPEN (a, k) -> print_string ("open "^a^"\n"); print_active_messages l
	  
(* cette fonction sert  remlacer un nom par la valeur associe dans
   l'environnement donn *)

and replace_from_env name = function
    [] -> Ident name
  | a::l -> if ((fst a)=name) then (snd a) else replace_from_env name l 

(* cette fonction rpond true si le premier lment dans une suite de
   capacits est un identificateur *)

and is_first_capa_ident = function
    Ident n -> true
  | Seq_capa (c1, c2) -> is_first_capa_ident c1
  | _ -> false
